import json
import hashlib
from types import SimpleNamespace


class TensorEncoder(json.JSONEncoder):
    fast_bytes = False

    def default(self, obj):
        original = obj
        if hasattr(obj, "__attrs_attrs__"):
            typename = type(obj).__name__
            obj = {a.name: getattr(obj, a.name) for a in obj.__attrs_attrs__}
            obj[".attr.__name__"] = typename
        if isinstance(obj, SimpleNamespace):
            obj = obj.__dict__
            obj[".builtin.__name__"] = "types.SimpleNamespace"
        obj = best_effort_to_bytes(obj)
        if isinstance(obj, bytes):
            if self.fast_bytes:
                obj = hashlib.sha256(obj).hexdigest()
            else:
                obj = str(obj)
        if isinstance(obj, range):
            obj = {".type": "range", "representation": str(obj)}
        if any(x.__name__ == "Module" for x in type(obj).mro()):
            obj = {
                ".type": "Module",
                "hash": dict(
                    other_dict={
                        k: v for k, v in obj.__dict__.items() if not k.startswith("_")
                    },
                    state_dict=obj.state_dict(),
                ),
            }
        if obj is original:
            return super().default(obj)
        return obj


class FastTensorEncoder(TensorEncoder):
    fast_bytes = True


def best_effort_to_bytes(obj):
    if type(obj).__module__ == "torch" and type(obj).__name__ == "Tensor":
        if obj.is_cuda:
            obj = obj.cpu()
        obj = obj.detach().numpy()
    if type(obj).__module__ == "numpy":
        if type(obj).__name__ == "dtype":
            obj = str(obj).encode("utf-8")
        else:
            obj = obj.tostring()
    return obj


def stringify(obj, *, fast_bytes=False):
    return json.dumps(
        obj, cls=FastTensorEncoder if fast_bytes else TensorEncoder, sort_keys=True
    )


def stable_hash(obj, *, fast_bytes=True):
    return hashlib.sha256(
        stringify(obj, fast_bytes=fast_bytes).encode("utf-8")
    ).hexdigest()



import os

from filelock import FileLock
import shelve
import time


class Lock:
    def __init__(self, lock_path, time_path):
        self.lock = FileLock(lock_path)
        self.time_path = time_path
        self.last_opened = float("-inf")
        self.unlocked = False

    def _get_last_modified(self):
        self._check()
        try:
            with open(self.time_path) as f:
                return float(f.read())
        except:
            self.set_last_modified()
            return self._get_last_modified()

    def opened_after_last_modification(self):
        self._check()
        last_modified_time = self._get_last_modified()
        return self.last_opened >= last_modified_time

    def set_last_opened(self):
        self.last_opened = time.time()

    def set_last_modified(self):
        self._check()
        self.set_last_opened()
        with open(self.time_path, "w") as f:
            f.write(str(self.last_opened))

    def __enter__(self):
        self.lock.__enter__()
        self.unlocked = True
        return self

    def __exit__(self, *args, **kwargs):
        self.unlocked = False
        self.lock.__exit__(*args, **kwargs)

    def _check(self):
        assert self.unlocked, "can only perform this operation on an unlocked lock"


class LockedShelf:
    """
    A class that manages a shelf that can be accessed from multiple threads simultaneously.

    Mantains a cache of some of the elements of the dictionary. If it ever was not
        the most recent accessor of the shelf, the cache is entirely flushed.

    The cache is mantained over opening and closing of the shelf.
    """

    def __init__(self, path):
        try:
            os.makedirs(path)
        except FileExistsError:
            pass
        self.path = path
        self.lock = Lock(self.path + "/lock", self.path + "/time")
        self.shelve_path = self.path + "/shelf"
        self.shelf = None
        self.cache = None

    def _update(self):
        if self.shelf is None:
            self.shelf = shelve.open(self.shelve_path)
            self.cache = {}
        else:
            assert self.cache is not None
            if not self.lock.opened_after_last_modification():
                self.cache = {}
                self.lock.set_last_opened()

    def __getitem__(self, key):
        self._update()

        if key not in self.cache:
            result = self.cache[key] = self.shelf[key]
            return result

        return self.cache[key]

    def __contains__(self, key):
        self._update()

        return key in self.cache or key in self.shelf

    def __setitem__(self, key, value):
        self._update()
        self.cache[key] = value
        self.shelf[key] = value
        self.lock.set_last_modified()

    def __enter__(self):
        self.lock.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self.lock.__exit__(*args, **kwargs)

    def close(self):
        self.shelf.close()



import os

import shutil

from appdirs import user_cache_dir

from .locked_shelf import LockedShelf
from .hash import stringify
from .dict_function import dict_function

CACHE = user_cache_dir("permacache")


class CachedFunction:
    def __init__(self, function, key_function, path):
        self.function = function
        self.key_function = key_function
        self.shelf = LockedShelf(path)

    def __call__(self, *args, **kwargs):
        key = stringify(self.key_function(*args, **kwargs))
        with self.shelf as db:
            if key in db:
                return db[key]
        value = self.function(*args, **kwargs)
        with self.shelf as db:
            # TODO maybe check if key is now in db
            db[key] = value
        return value


def permacache(path, key_function=dict()):
    path = os.path.join(CACHE, path)

    def annotator(f):
        kf = key_function
        if isinstance(kf, dict):
            kf = dict_function(kf, f)
        return CachedFunction(f, kf, path)

    return annotator


def to_file(path, zip_path):
    path = os.path.join(CACHE, path)
    shutil.make_archive(zip_path, "zip", path)


def from_file(path, zip_path):
    path = os.path.join(CACHE, path)
    if os.path.exists(path):
        raise RuntimeError(f"Cache already exists: {path}")
    shutil.unpack_archive(zip_path + ".zip", path, "zip")



import inspect


DEFAULT_FUNCTIONS = {None: lambda x: None}


class drop_if:
    def __init__(self, predicate, mapper=lambda x: x):
        self.predicate = predicate
        self.mapper = mapper


def drop_if_equal(value, mapper=lambda x: x):
    return drop_if(lambda x: x == value, mapper)


def dict_function(d, fn):
    signature = inspect.signature(fn)

    def key(*args, **kwargs):
        arguments = signature.bind(*args, **kwargs)
        arguments.apply_defaults()
        result = {}
        for k, v in arguments.arguments.items():
            if k in d:
                if isinstance(d[k], drop_if):
                    if d[k].predicate(v):
                        continue
                    v = d[k].mapper(v)
                elif callable(d[k]):
                    v = d[k](v)
                else:
                    v = DEFAULT_FUNCTIONS[d[k]](v)
            result[k] = v
        return result

    return key
